﻿using System;
using System.Collections.Generic;
using System.Text;

namespace OpenDelegateToInstanceMethod {
	class Program {

		static int Add(int a, int b) => a + b;

		static int Multiply(int a, int b) => a * b;

		static void Main(string[] args) {
			//
			// Open delegates = no target -> to static methods
			//

			var ops = new List<Func<int, int, int>> { Add, Multiply };

			foreach (var op in ops) {
				Console.WriteLine($"op result: {op(10, 20)}");
			}

			var str = "Homeworld";

			//
			// Closed delegates = target to "this" -> to instance methods
			//
			var finders = new List<Func<char, int>> { str.IndexOf, str.LastIndexOf };

			Console.WriteLine($"for \"{str}\":");
			foreach (var finder in finders) {
				Console.WriteLine($"finder result: {finder('o')}");
			}

			str = "Closure";

			Console.WriteLine($"for \"{str}\":");
			foreach (var finder in finders) {
				Console.WriteLine($"finder result: {finder('o')}");
			}

			Console.WriteLine();

			//
			// Open delegates = no target -> to instance methods (no C# syntax, have to use Reflection)
			//
			var openFinders = new List<Func<string, char, int>> {
				
				(Func<string, char, int>) Delegate.CreateDelegate(
					typeof(Func<string, char, int>),
					typeof(string).GetMethod("IndexOf", new [] { typeof(char) })
				),

				(Func<string, char, int>) Delegate.CreateDelegate(
					typeof(Func<string, char, int>),
					typeof(string).GetMethod("LastIndexOf", new [] { typeof(char) })
				),
			};

			str = "Homeworld";
			foreach (var openFinder in openFinders) {
				Console.WriteLine($"openFinder result: {openFinder(str, 'o')}");
			}
			str = "Closure";
			foreach (var openFinder in openFinders) {
				Console.WriteLine($"openFinder result: {openFinder(str, 'o')}");
			}

			//
			// Closed delegates = target set -> to static N parameter method = target is passed to 1st argument, delegate created for remaining N-1 arguments
			//
			
			// NOT SUPPORTED FOR 1ST PARAMETER BEING A VALUE TYPE:
			// Throws System.ArgumentException: Cannot bind to the target method because its signature is not compatible with that of the delegate type.

			//var closedOp = (Func<int, int>) Delegate.CreateDelegate(    
			//	typeof(Func<int, int>),
			//	5,
			//	typeof(Program).GetMethod("Add", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Static)
			//);

			//Console.WriteLine(closedOp(10));
			//Console.WriteLine(closedOp(20));

			// But works for reference types:
			var closedHelloOp = (Func<string, string>) Delegate.CreateDelegate(
				typeof(Func<string, string>),
				"Hello ",
				typeof(Program).GetMethod("Concat", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Static)
			);

			Console.WriteLine(closedHelloOp("world"));
			Console.WriteLine(closedHelloOp("Pavel"));

			Console.WriteLine();

			//
			// This "partial function application" can be done via lambdas in a simpler manner:
			//

			Func<string, string> helloOp = static s2 => Concat("Hello ", s2);
			Console.WriteLine(helloOp("world"));
			Console.WriteLine(helloOp("Pavel"));

			Func<int, int> fiveAdd = static b => Add(5, b);
			Console.WriteLine(fiveAdd(10));
			Console.WriteLine(fiveAdd(20));
		}

		static string Concat(string s1, string s2) => s1 + s2;
	}
}
